class GeneticDiversity(Exception):
    """
    Exception raised when the mango_genetic diversity of the population is too low.
    """

    pass


class ConfigurationError(Exception):
    """
    Exception raised when the configuration of the mango_genetic algorithm is not valid.
    """

    pass
